"""
input初期的测试代码，放到这只做备忘
"""

import sys
from terminal_layout import KeyListener, Key
import logging

from terminal_layout.types import String

logger = logging.getLogger(__name__)
logger.setLevel(level=logging.DEBUG)
handler = logging.FileHandler("terminal_layout.log")
logger.addHandler(handler)

from terminal_layout.ansi import Cursor


k = KeyListener()

s = String('')
s_char_list_index = 0
cursor_index = 0


@k.bind_key('any')
def _(_, e):
    global s, s_char_list_index, cursor_index

    c = e.key.code
    logger.debug(f's={s},  c={c}')

    if repr(c).startswith("'\\x") or repr(c).startswith("u'\\x"):

        if e.key == Key.LEFT and s_char_list_index > 0:
            back = len(s.char_list[s_char_list_index - 1])
            cursor_index -= back
            s_char_list_index -= 1
            sys.stdout.write(Cursor.BACK(back))

        if e.key == Key.RIGHT and s_char_list_index < len(s.char_list):
            forward = len(s.char_list[s_char_list_index])
            cursor_index += forward
            s_char_list_index += 1
            sys.stdout.write(Cursor.FORWARD(forward))

        if e.key == Key.BACKSPACE and s_char_list_index > 0:
            back = len(s.char_list[s_char_list_index - 1])
            cursor_index -= back
            s_char_list_index -= 1

            len_right_s = 0
            show_s = ''
            for c in s.get_char_list_item(s_char_list_index + 1):
                len_right_s += len(c)
                show_s += str(c)

            s.pop(s_char_list_index)

            logger.info(f'{back} - {len_right_s} - {show_s} - {repr(Cursor.BACK(len_right_s) if len_right_s else "")}')

            sys.stdout.write(
                Cursor.BACK(back) + ' ' * (len_right_s + back) + Cursor.BACK(len_right_s + back) + show_s)

            if len_right_s > 0:
                sys.stdout.write(Cursor.BACK(len_right_s))


    else:
        show_s = c
        c_str = String(c)
        after_show_s = ''
        len_right_s = 0
        for c in s.get_char_list_item(s_char_list_index):
            len_right_s += len(c)
            show_s += str(c)

        if len_right_s > 0:
            sys.stdout.write(' ' * len_right_s + Cursor.BACK(len_right_s))
            after_show_s = Cursor.BACK(len_right_s)

        sys.stdout.write(show_s + after_show_s)

        s.insert_into_char_list(s_char_list_index, c_str)
        s_char_list_index += len(c_str.char_list)
        cursor_index += len(c_str)
        logger.debug(f's_char_list_index={s_char_list_index}, cursor_index={cursor_index}, s={s}')

    sys.stdout.flush()


s1 = 'bbb, input:'
sys.stdout.write(s1)
sys.stdout.write('\n')
sys.stdout.write('asd222\n' + Cursor.UP(2) + Cursor.FORWARD(len(s1)))
sys.stdout.flush()
k.listen(stop_key=[Key.ENTER])

print('\nasd222')



import sys
import inspect

"""
instance_variables

-----------------------------         ------------------------------                        
|def __init__(self, a, b):  |         | @instance_variables        |
|    self.a=a               |   ==>   | def __init__(self, a, b):  |
|    self.b=b               |         |     pass                   |
-----------------------------         ------------------------------ 

"""
if sys.version_info >= (3, 0):
    from functools import wraps


    def instance_variables(f):
        sig = inspect.signature(f)

        @wraps(f)
        def wrapper(self, *args, **kwargs):
            values = sig.bind(self, *args, **kwargs)
            for k, p in sig.parameters.items():
                if k != 'self':
                    if k in values.arguments:
                        val = values.arguments[k]
                        if p.kind in (inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.KEYWORD_ONLY):
                            setattr(self, k, val)
                        elif p.kind == inspect.Parameter.VAR_KEYWORD:
                            for k, v in values.arguments[k].items():
                                setattr(self, k, v)
                    else:
                        setattr(self, k, p.default)
            
            f(self, *args, **kwargs)
        return wrapper
else:
    def instance_variables(func):
        def return_func(*args, **kwargs):
            self_var = args[0]

            arg_spec = inspect.getargspec(func)
            argument_names = arg_spec[0][1:]
            defaults = arg_spec[3]
            if defaults is not None:
                default_arg_dict = dict(zip(reversed(argument_names), reversed(defaults)))
                self_var.__dict__.update(default_arg_dict)

            arg_dict = dict(zip(argument_names, args[1:]))
            self_var.__dict__.update(arg_dict)

            valid_keywords = set(kwargs) & set(argument_names)
            kwarg_dict = {k: kwargs[k] for k in valid_keywords}
            self_var.__dict__.update(kwarg_dict)

            func(*args, **kwargs)

        return return_func

# 这个应该是py2的代码，不知道之前为啥统一用了这个
# def instance_variables(func):
#     def return_func(*args, **kwargs):
#         self_var = args[0]

#         arg_spec = inspect.getargspec(func)
#         argument_names = arg_spec[0][1:]
#         defaults = arg_spec[3]
#         if defaults is not None:
#             default_arg_dict = dict(zip(reversed(argument_names), reversed(defaults)))
#             self_var.__dict__.update(default_arg_dict)

#         arg_dict = dict(zip(argument_names, args[1:]))
#         self_var.__dict__.update(arg_dict)

#         valid_keywords = set(kwargs) & set(argument_names)
#         kwarg_dict = {k: kwargs[k] for k in valid_keywords}
#         self_var.__dict__.update(kwarg_dict)

#         func(*args, **kwargs)

#     return return_func



"""
inputView两种实现方法：
1. 外部监听key事件，然后通过 setText 修改已有的textView值。这样做需要自己移动光标，频繁的redraw
2. 基于textView写个view，这样做需要在view内自己处理 宽度，以及父view 相关的逻辑问题

InputEx使用第一种
"""
import sys

from terminal_layout.log import logger
from terminal_layout.ctl import TextViewProxy, LayoutProxy

try:
    from typing import Union
except:
    pass
from terminal_layout import *

from terminal_layout.types import String


class InputEx(object):
    view = None  # type:Union[None, TextViewProxy, LayoutProxy]
    x = -1
    y = -1
    layout_height = 0
    input_s = None  # type: String
    input_char_list_index = 0
    input_char_list_start = 0
    input_char_list_end = 0
    cursor_index = 0
    max_length = None

    def __init__(self, _ctl, input_buffer=30):
        """

        :param input_buffer: 读取字符的缓存；输入法一次输入多个字符时，若buffer过小会导致只读取到一部分

        :type _ctl:LayoutCtl
        """
        self.ctl = _ctl
        self.input_buffer = input_buffer

    def get_view_y(self,id,row):
        y=0
        for v in row.data:
            if v.id == id:
                if v.visibility != Visibility.visible:
                    return True, -1
                return True, y
            y += v.real_width or 0
        return False,-1


    def get_view_x_y(self, id):
        """

        (x, y)--------------
        |   1     2     3   |
        |   4     5     6   |
        |   7     8     9   |
        --------------(x, y)

        (1,2) = 6
        """
        x = 0
        y = 0
        layout=self.ctl.layout
        if isinstance(layout,TextView):
            return -1,-1 if layout.id!=id else x,y
        if isinstance(layout,TableRow):
            ok,y=self.get_view_y(id,layout)
            return x,y
        for row in self.ctl.layout.data:
            if row.visibility == Visibility.gone:
                continue
            if row.visibility == Visibility.visible:
                ok,y=self.get_view_y(id,row)
                if ok:
                    return x,y
            x += 1
        return -1, -1

    def get_parent_max_width(self, view):
        p = view.get_parent()
        if p is None:
            return self.ctl.get_terminal_size()[0]
        else:
            parent_max_width = self.get_parent_max_width(p)
            width = p.get_width()
            if isinstance(width, int) and width < parent_max_width:
                return width
            return parent_max_width

    def get_view_max_width(self):
        parent_max_width = self.get_parent_max_width(self.view)

        parent = self.view.get_parent()
        if parent is not None:
            for v in parent.data:
                if v.id == self.view.get_id():
                    continue
                if v.weight is None:
                    parent_max_width -= v.real_width

        width = self.view.get_width()

        if isinstance(width, int) and width < parent_max_width:
            return width
        return parent_max_width

    def get_input(self, id, max_length=None):
        """
        :param max_length: 输入字符数；注意区分max_width，max_width表示显示宽度，和输入无关
        """
        self.view = self.ctl.find_view_by_id(id)
        self.max_length = max_length

        if not self.view or not isinstance(self.view, TextViewProxy):
            return False, '-1'

        self.x, self.y = self.get_view_x_y(id)
        if self.x < 0 or self.y < 0:
            return False, '-2'

        # 停止绘制线程
        if self.ctl.auto_re_draw:
            self.ctl.refresh_thread_stop = True
            self.ctl.refresh_thread.join()

        self.max_width = self.get_view_max_width()

        if self.max_width <= 0:
            return False, '-3'

        self.layout_height = self.ctl.layout.real_height

        self.input_s = String(self.view.get_text())
        if self.max_length and len(self.input_s.char_list) > self.max_length:
            self.input_s.char_list_slice(stop=self.max_length)

        self.input_char_list_index = len(self.input_s.char_list)
        self.input_char_list_start = 0
        self.input_char_list_end = len(self.input_s.char_list)
        self.move_cursor_to_view()
        self.update_index_and_show_s()

        kl = KeyListener()

        kl.bind_key('any', self.key_event, decorator=False)

        kl.listen()
        self.move_cursor_to_btm()

        if self.ctl.auto_re_draw:
            self.ctl.init_refresh_thread()
            self.ctl.refresh_thread.start()

        return True,str(self.input_s)

    def move_cursor_to_view(self):
        s = ''
        if self.layout_height - self.x > 0:
            s += Cursor.UP(self.layout_height - self.x)
        if self.y + self.cursor_index > 0:
            s += Cursor.FORWARD(self.y + self.cursor_index)
        if s:
            sys.stdout.write(s)
            sys.stdout.flush()

    def char_list_iter(self, i, reverse):
        for i in range(i - 1, -1, -1) if reverse else range(i, len(self.input_s.char_list), 1):
            yield i, self.input_s.char_list[i]

    def update_index_and_show_s(self, right=True):
        """
        right: 光标从左向右移动
        """

        if len(self.input_s) >= self.max_width:
            left_width = self.max_width
            if right:
                list_iter = self.char_list_iter(self.input_char_list_end, True)
            else:
                list_iter = self.char_list_iter(self.input_char_list_index, False)

            new_i = 0
            for i, c in list_iter:

                if left_width - len(c) < 0:
                    break
                left_width -= len(c)
                new_i = i

            if right:
                self.input_char_list_start = new_i
            else:
                self.input_char_list_end = new_i + 1

        self.cursor_index = 0
        show_s = ''
        for i in range(self.input_char_list_start, self.input_char_list_end):
            show_s += str(self.input_s.char_list[i])
            if i < self.input_char_list_index:
                self.cursor_index += len(self.input_s.char_list[i])

        self.move_cursor_to_btm()
        self.view.set_text(show_s)
        self.ctl.re_draw()
        self.move_cursor_to_view()

    def move_cursor_to_btm(self):
        s = ''
        if self.layout_height - self.x > 0:
            s += Cursor.DOWN(self.layout_height - self.x)
        if self.y + self.cursor_index > 0:
            s += Cursor.BACK(self.y + self.cursor_index + 2)
        if s:
            sys.stdout.write(s)
            sys.stdout.flush()

    def update_text(self):

        self.view.set_text(str(self.input_s))
        self.ctl.update_width()

    def key_event(self, kl, event):
        c = event.key.code
        right = True

        if len(repr(c)) > 1 and (repr(c).startswith("'\\") or repr(c).startswith("u'\\")):
            if event.key == Key.LEFT and self.input_char_list_index > 0:

                self.input_char_list_index -= 1
                if self.input_char_list_index < self.input_char_list_start:
                    self.input_char_list_start = self.input_char_list_index
                    right = False

            elif event.key == Key.RIGHT and self.input_char_list_index < len(self.input_s.char_list):

                self.input_char_list_index += 1
                if self.input_char_list_index > self.input_char_list_end - 1:
                    self.input_char_list_end = min(self.input_char_list_index + 1, len(self.input_s.char_list))

            elif event.key == Key.BACKSPACE and self.input_char_list_index > 0:

                self.input_char_list_index -= 1
                self.input_s.pop(self.input_char_list_index)
                self.input_char_list_end -= 1

            elif event.key == Key.ENTER:
                kl.stop()
                return
        else:

            c_str = String(c)
            if self.max_length:
                left_length = self.max_length - len(self.input_s.char_list)
                if left_length == 0:
                    return
                if len(c_str.char_list) > left_length:
                    c_str.char_list_slice(stop=left_length)

            self.input_s.insert_into_char_list(self.input_char_list_index, c_str)
            self.input_char_list_index += len(c_str.char_list)
            self.input_char_list_end += len(c_str.char_list)

        self.update_index_and_show_s(right=right)



from terminal_layout import *
from terminal_layout.helper.class_helper import instance_variables
from terminal_layout.extensions.scroll import *


class StringStyle(object):

    def __init__(self, fore=None, back=None, style=None):
        self.fore = fore or ''
        self.back = back or ''
        self.style = style or ''

    def to_dict(self):
        return {
            'fore': self.fore,
            'back': self.back,
            'style': self.style
        }


class Choice(object):
    @instance_variables
    def __init__(self, title, choices, icon='> ', icon_style=StringStyle(fore=Fore.green), choices_style=StringStyle(),
                 selected_style=StringStyle(), loop=True, default_index=0, stop_key=None):
        self.current = self.get_default_index(default_index)
        self.result = None
        if not stop_key:
            self.stop_key = ['q']

    def get_default_index(self, default_index):
        if default_index < 0 or default_index > len(self.choices):
            return 0
        return default_index

    h_cache = None

    def hidden_choices(self):
        """
        适配高度，当高度不够时隐藏部分choices
        """
        if self.h_cache is None:
            _, self.h_cache = self.ctl.get_terminal_size()
            # 最后一行要留着显示光标，不能输出因此-1
            # title不能隐藏，因此再-1
            self.h_cache -= 2

        if self.h_cache >= len(self.choices) + 1:
            return

        # 优先显示current下面部分，还有剩余高度时从current上面补
        show_start = self.current
        show_end = len(self.choices)
        if show_end - show_start > self.h_cache:
            show_end = show_start + self.h_cache
        left = self.h_cache - show_end + show_start
        if left > 0:
            show_start -= left
        row_id_p = 'root_row_'
        for i in range(len(self.choices)):
            if i >= show_start and i < show_end:
                self.ctl.find_view_by_id(
                    row_id_p + str(i + 1)).set_visibility(Visibility.visible)
            else:
                self.ctl.find_view_by_id(
                    row_id_p + str(i + 1)).set_visibility(Visibility.gone)

    def get_choice(self):
        views = [[TextView('', self.title)]]
        for i, c in enumerate(self.choices):
            views.append(
                [TextView('icon%d' % i, self.icon, visibility=Visibility.invisible, **self.icon_style.to_dict()),
                 TextView('value%d' % i, c, **self.choices_style.to_dict())])
        views[self.current + 1][0].visibility = Visibility.visible
        views[self.current + 1][1] = TextView('value%d' % self.current, self.choices[self.current],
                                              **self.selected_style.to_dict())
        self.ctl = LayoutCtl.quick(TableLayout, views)
        # self.hidden_choices()
        # self.ctl.draw(auto_re_draw=False)

        self.scroll = Scroll(self.ctl, scroll_box_start=1, default_scroll_start=self.default_index,
                             up_key=None, down_key=None, stop_key=None, btm_text='')

        kl = self.scroll.init_kl()
        self.scroll.draw()
        kl.bind_key(Key.UP, Key.DOWN, self.change_current, decorator=False)
        kl.bind_key(Key.ENTER, self.select, decorator=False)

        kl.listen(self.stop_key)
        return self.result

    def update_style(self, view, style):
        view.set_back(style.back)
        view.set_style(style.style)
        view.set_fore(style.fore)

    def change_current(self, kl, event):
        temp = self.current
        loop_tigger = False
        if event.key == Key.UP:
            temp -= 1
        elif event.key == Key.DOWN:
            temp += 1
        if temp < 0:
            if self.loop:
                loop_tigger = True
                temp = len(self.choices) - 1
            else:
                temp = 0
        if temp >= len(self.choices):
            if self.loop:
                loop_tigger = True
                temp = 0
            else:
                temp = len(self.choices) - 1
        self.ctl.find_view_by_id(
            'icon%d' % self.current).set_visibility(Visibility.invisible)
        self.update_style(self.ctl.find_view_by_id('value%d' %
                                                   self.current), self.choices_style)

        self.current = temp
        self.ctl.find_view_by_id(
            'icon%d' % self.current).set_visibility(Visibility.visible)
        self.update_style(self.ctl.find_view_by_id('value%d' %
                                                   self.current), self.selected_style)
        # self.hidden_choices()
        self.scroll.loop = loop_tigger
        if event.key == Key.UP:
            self.scroll.up()
        else:
            self.scroll.down()
        self.ctl.re_draw()

    def select(self, kl, event):
        self.result = (self.current, self.choices[self.current])
        self.scroll.stop(kl)



# -*- coding: utf-8 -*-
import copy

char_to_length = {
    '\t': 4,
}
char_to_str = {
    '\t': ' ' * 4,
    '\n': ' ',
}


def add_char_length_map(d):
    """
    ex:
    add_char_length_map({'a':1})
    :param d: 
    :type d: dict
    """
    char_to_length.update(d)


class Char(object):
    length = None

    def __init__(self, c):
        if len(c) != 1:
            raise TypeError(
                'expected a character, but string of length %d found' % len(c))
        self.c = c

    def __len__(self):
        if self.length is not None:
            return self.length

        temp = char_to_length.get(self.c, None)
        if temp:
            self.length = temp
            return self.length

        if ord(self.c) < 11904:
            self.length = 1
        else:
            self.length = 2
        return self.length

    def __str__(self):
        return char_to_str.get(self.c, self.c)


class String(object):
    """

    ex:

    >>> s=String('a啊啊')
    >>> len(s)
    5
    >>> s[:2]
    'a'
    >>> s[:3]
    'a啊'

    """
    length = None
    char_list = None

    def __init__(self, text):
        """

        :param text:
        :type text:str
        """
        self.char_list = []  # type: list[Char]
        self.origin_text = text
        for c in text:
            self.char_list.append(Char(c))

    def __len__(self):
        if self.length is not None:
            return self.length

        t = 0
        for c in self.char_list:
            t += len(c)
        self.length = t

        return self.length

    def __str__(self):
        return ''.join(str(c) for c in self.char_list)

    def __add__(self, other):
        new_str = copy.deepcopy(self)
        new_str.length = None
        if isinstance(other, str):
            for c in other:
                new_str.char_list.append(Char(c))
            new_str.origin_text += other
        if isinstance(other, String):
            new_str.char_list += other.char_list
            new_str.origin_text += other.origin_text

        if isinstance(other, Char):
            new_str.char_list.append(other)
            new_str.origin_text += str(other)
        return new_str

    def __radd__(self, other):
        new_str = copy.deepcopy(self)
        new_str.length = None
        if isinstance(other, str):
            for c in other:
                new_str.char_list.insert(0, Char(c))
            new_str.origin_text = other + new_str.origin_text

        if isinstance(other, Char):
            new_str.char_list.insert(0, other)
            new_str.origin_text = str(other) + new_str.origin_text

        return new_str

    def __getitem__(self, item):
        """
        >>> s=String('a啊啊')
        >>> s[:2]
        'a'
        >>> s[:3]
        'a啊'

        :rtype: str
        """
        start = item.start or 0
        if start != 0:
            raise TypeError('slice start must be 0 or None')
        stop = item.stop
        if stop is None or stop < 0:
            raise TypeError('slice stop must be positive integer')

        s = ''
        for c in self.char_list:
            length = len(c)
            if stop >= length:
                s += str(c)
            else:
                break
            stop -= length

        return s

    def insert_into_char_list(self, i, s):

        if isinstance(s, str):
            s = String(s).char_list
        if isinstance(s, Char):
            s = [s]
        if isinstance(s, String):
            s = s.char_list
        self.char_list = self.char_list[:i] + s + self.char_list[i:]
        self.origin_text = str(self)
        self.length = None

    def get_char_list_item(self, start=None, stop=None):
        """
        >>> s=String('123')
        >>> s.get_char_list_item() # s.char_list[:]
        ['1','2','3']
        >>> s.get_char_list_item(start=1) # s.char_list[1:]
        ['2','3']

        """
        return self.char_list[start:stop]

    def pop(self, i=-1):
        self.char_list.pop(i)
        self.origin_text = str(self)
        self.length = None

    def char_list_slice(self, start=None, stop=None):
        self.char_list = self.get_char_list_item(start, stop)
        self.origin_text = str(self)
        self.length = None



import logging

logger = logging.getLogger(__name__)
logger.setLevel(level=logging.DEBUG)


def disable_logger():
    handler = logging.NullHandler()
    logger.handlers = [handler]


def enable_logger():
    handler = logging.FileHandler("terminal_layout.log")
    handler.setLevel(logging.DEBUG)
    # formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    formatter = logging.Formatter('%(levelname)s : %(message)s')

    handler.setFormatter(formatter)
    logger.addHandler(handler)


disable_logger()



"""

"""
from terminal_layout.helper.class_helper import instance_variables
from terminal_layout.view.params import OverflowVertical, Visibility
from terminal_layout.view.layout import TableRow
from terminal_layout.view.text_view import TextView
from terminal_layout.readkey import Key, KeyListener
from terminal_layout.log import logger
from terminal_layout.helper.helper import get_terminal_size


class ScrollEvent:
    up = 'up'
    down = 'down'
    stop = 'stop'


scroll_btm_row = '_scroll_btm_row_'
scroll_btm_text = '_scroll_btm_text_'


class Scroll(object):
    """
    注意：只支持TableLayout ！！！
    """
    __slots__ = (
        'ctl', 'stop_key', 'up_key', 'down_key', 'scroll_box_start', 'default_scroll_start', 'loop', 'btm_text', 'more',
        'callback', 're_draw_after_scroll', 're_draw_after_stop', 'old_layout_overflow', 'current_scroll_start',
        'key_listener')

    @instance_variables
    def __init__(self, ctl,
                 stop_key=Key.ESC, up_key=Key.UP, down_key=Key.DOWN,
                 scroll_box_start=0, default_scroll_start=0, loop=False,
                 btm_text='', more=False,
                 callback=None, re_draw_after_scroll=True, re_draw_after_stop=False):
        """
        :param ctl: ctl
        :param overflow: hidden or scroll

        :param stop_key: 停止scroll的key
        :param up_key:
        :param down_key:
        :param scroll_box_start: 从哪行开始可以滚动。若第一行要显示标题，则scroll_start=1
        :param default_scroll_start: 初始化时, scroll_box中哪行显示在第一的位置。
        :param loop: 循环
        :param btm_text: 底部的文本，为空则不显示
        :param more: 类似于man的效果。Ture会自动添加 btm_text
        :param callback: 滚动后的回调
        :param re_draw_after_scroll: 滚动后执行重绘。为false时你需要自己调用re_draw
        :param re_draw_after_stop:


        default_scroll_start 说明：

        若terminal高度为4, table有6行（即高度为6），default_scroll_start=6。
        绘制时，显示在terminal顶部的是row_2。

           row_0
           row_1
         |------------|
         | row_2      |
         | row_3      |  <=== terminal, h=4
         | row_4      |
         | row_5      |
         |------------|


        """
        self.old_layout_overflow = None
        self.current_scroll_start = default_scroll_start
        if self.current_scroll_start < self.scroll_box_start:
            self.current_scroll_start = self.scroll_box_start

        """
        current_scroll_start : 当前实际显示的滚动区域的第一行, 详细说明如下

            
          row_0
        |------------------------|
        | row_1                  |  <= terminal, height=4
        | row_2                  |
        | row_3                  |
        | row_4                  |
        |------------------------|

        外框表示terminal，其高度为4。table高度为5 row 1-4 显示在terminal中。
        此时current_scroll_start = 1
        若 loop=False ，触发down时 current_scroll_start不变 （ 注意不会+1 !!）
        若 loop=True ，触发down时 current_scroll_start=0
        """
        if more:
            self.btm_text = '-- more --'

    def init_kl(self, stop_func=None, up_func=None, down_func=None):
        """
        :rtype: KeyListener
        """
        self.key_listener = KeyListener()
        if self.stop_key:
            self.key_listener.bind_key(self.stop_key, stop_func or self._stop, decorator=False)
        if self.up_key:
            self.key_listener.bind_key(self.up_key, up_func or self._up, decorator=False)
        if self.down_key:
            self.key_listener.bind_key(self.down_key, down_func or self._down, decorator=False)
        return self.key_listener

    def draw(self):
        # 修改table的overflow
        table = self.ctl.get_layout()
        self.old_layout_overflow = table.get_overflow_vertical()
        table.set_overflow_vertical(OverflowVertical.none)
        # 添加btm
        if self.btm_text:
            r = TableRow.quick_init(scroll_btm_row,
                                    TextView(scroll_btm_text, self.btm_text)
                                    )
            table.add_view(r)

        self.cal_and_update(self.current_scroll_start, is_first=True)

        self.ctl.draw(auto_re_draw=False)

    def scroll(self, stop_func=None, up_func=None, down_func=None):
        self.init_kl(stop_func, up_func, down_func)
        self.draw()
        self.key_listener.listen()

    def _callback(self, event):
        if not self.callback:
            return
        self.callback(event)

    def _up(self, kl, event):
        self.up()
        if self.re_draw_after_scroll:
            self.ctl.re_draw()
        self._callback(ScrollEvent.up)

    def up(self):
        self.cal_and_update(self.current_scroll_start - 1)

    def _down(self, kl, event):
        self.down()
        if self.re_draw_after_scroll:
            self.ctl.re_draw()
        self._callback(ScrollEvent.down)

    def down(self):
        self.cal_and_update(self.current_scroll_start + 1)

    def _stop(self, kl, event):
        self.stop(kl)
        if self.re_draw_after_stop:
            self.ctl.draw()
        self._callback(ScrollEvent.stop)

    def stop(self, kl):
        kl.stop()
        table = self.ctl.get_layout()
        table.set_overflow_vertical(self.old_layout_overflow)
        if self.btm_text:
            table.remove_view_by_id('_scroll_btm_row_')
            table.view.old_row_visibility.pop(-1)

    def cal_and_update(self, new_current_scroll_start, is_first=False):
        """
        cal_scroll 和 update_view最开始是在一起的，
        之所以拆成两个函数是为了满足需要cal_scroll的计算结果的需求
        """
        r = self.cal_scroll(new_current_scroll_start, is_first)
        # 等于0说明高度足够，不用修改view的visibility
        if r[4] != 0:
            self.update_view(*r)

    def cal_scroll(self, new_current_scroll_start, is_first=False):
        _, h = self.ctl.get_terminal_size()
        # 最后一行需要显示光标，因此-1
        h -= 1 + self.scroll_box_start

        table = self.ctl.get_layout().view

        scroll_box_end = len(table.data)

        if self.btm_text:
            h -= 1
            scroll_box_end -= 1

        scroll_box_h = scroll_box_end - self.scroll_box_start

        for r in table.data:
            if r.visibility == Visibility.gone:
                scroll_box_h -= 1
        if h >= scroll_box_h:
            return 0, 0, 0, 0, 0

        scroll_start = new_current_scroll_start
        if scroll_start < self.scroll_box_start:
            # 首次绘制时，不会出现 scroll_start<self.scroll_box_start
            # 因此进这里一定是按了up键
            if self.loop:
                scroll_start = scroll_box_end - h
            else:
                scroll_start = self.scroll_box_start
        elif scroll_start + h > scroll_box_end:
            # 进这里说明是按down键，或者首次绘制时
            if self.loop and not is_first:
                scroll_start = self.scroll_box_start

        # 计算scroll_start， scroll_end
        # 优先显示scroll_start下面部分，高度还有剩余就从上面补

        left_h = h
        scroll_end = scroll_start

        # 因为可能有view隐藏的情况，这里直接循环到scroll_box_end，而不是 scroll_start+h
        for i, r in enumerate(table.data[scroll_start:scroll_box_end]):
            if not left_h:
                break

            scroll_end = scroll_start + i + 1
            if r.visibility != Visibility.gone:
                left_h -= 1

        while left_h and scroll_start - 1 >= self.scroll_box_start:
            scroll_start -= 1
            if table.data[scroll_start].visibility != Visibility.gone:
                left_h -= 1

        return self.scroll_box_start, scroll_box_end, scroll_start, scroll_end, h

    def update_view(self, scroll_box_start, scroll_box_end, scroll_start, scroll_end, h):
        self.current_scroll_start = scroll_start

        table = self.ctl.get_layout().view

        # 修改 visibility
        table.old_row_visibility = []
        for r in table.data[:scroll_box_start]:
            table.old_row_visibility.append(r.visibility)

        for i, r in enumerate(table.data[scroll_box_start:scroll_box_end]):
            i += scroll_box_start
            table.old_row_visibility.append(r.visibility)
            if i >= scroll_start and i < scroll_end:
                r._set_is_show(True)
            else:
                r.visibility = Visibility.gone
                r._set_is_show(False)

        if self.btm_text:
            btm = table.data[-1]
            table.old_row_visibility.append(btm.visibility)
            if self.more and scroll_end >= scroll_box_end:
                btm.data[0].text = '-- end --'
            else:
                btm.data[0].text = '-- more --'



# -*- coding: utf-8 -*-
from terminal_layout.view.params import Visibility, Gravity


class View(object):
    __slots__ = ('id', 'width', 'height', 'visibility', 'gravity',
                 'real_width', 'real_height', 'data', 'parent', '_is_show')

    def __init__(self, id, width, height=1, visibility=Visibility.visible, gravity=Gravity.left):
        """

        :param id:
        :param width:
        :param height: no used

        :type id:str
        :type width:
        :type height:int
        :type visibility:str
        :type gravity:str
        """

        self.data = []  # type: list[View]
        self.id = id
        self.width = width
        self.height = height
        self.visibility = visibility
        self.gravity = gravity

        self.real_width = None
        self.real_height = None
        self.parent = None

        # 是否显示在屏幕上。这个值用于： 当终端高度不够触发隐藏时，标识view是否显示在屏幕上。
        # 注意：你不应当主动修改这个值这是没意义的；另外is_show不会因visibility改变而改变。
        # 它只有下面两种情况才会被修改:
        # 1. tableLayout的overflow_vertical设为hidden_btm或hidden_top
        # 2. 使用scroll
        self._is_show = True

    def draw(self):
        pass

    def clear(self):
        pass

    def update_width(self, parent_width):
        pass

    def find_view_by_id(self, id):
        """

        :param id:
        :return:
        :rtype: View
        """
        if self.id == id:
            return self
        temp = None
        for v in self.data:
            # 这里不用判断v类型，textView的data是空list
            temp = v.find_view_by_id(id)
            if temp:
                break
        return temp

    def remove(self):
        """
        :rtype: bool
        """
        # 不能移除最外层view
        if not self.parent:
            return False
        return self.parent.remove_view_by_id(self.id)

    def remove_view_by_id(self, id):
        """
        :rtype: bool
        """
        if self.id == id:
            return self.remove()
        index = -1
        for i, v in enumerate(self.data):
            if v.id == id:
                index = i
                break
            # 这里不用判断v类型，textView的data是空list
            ok = v.remove_view_by_id(id)
            if ok:
                return ok
        if index == -1:
            return False
        self.data = self.data[:index]+self.data[index+1:]
        return True

    def get_width(self):
        """
        the width you set in __init__()
        初始化时设置的宽度

        :return:
        """
        return self.width

    def get_real_width(self):
        """

        the real width of the display
        实际显示的宽度

        :return:
        """
        return self.real_width

    def __getitem__(self, item):
        """

        :param item:
        :return:
        :rtype:View
        """
        return self.data[item]

    def insert(self, index, view):
        self.data.insert(index, view)

    def is_show(self):
        return self._is_show and self.visibility != Visibility.gone

    def _set_is_show(self, is_show):
        self._is_show = is_show

    terminal_w = 0
    terminal_h = 0

    def set_terminal_size(self, w, h):
        self.terminal_w = w
        self.terminal_h = h
